from LTO.AccountFactory import AccountFactory
from LTO import crypto
import struct
from nacl.signing import SigningKey, VerifyKey
import base58
from LTO.Account import Account

class AccountED25519(AccountFactory):

    def __init__(self, chainId):
        super().__init__(chainId)


    def createSignKeys(self, seed, nonce=0):
        seedHash = crypto.hashChain(struct.pack(">L", nonce) + crypto.str2bytes(seed))
        accountSeedHash = crypto.sha256(seedHash)
        privateKey = SigningKey(accountSeedHash)
        publicKey = privateKey.verify_key
        return privateKey, publicKey

    def createAddress(self, publicKey):
        unhashedAddress = chr(1) + str(self.chainId) + crypto.hashChain(publicKey.__bytes__())[0:20]
        addressHash = crypto.hashChain(crypto.str2bytes(unhashedAddress))[0:4]
        return base58.b58encode(crypto.str2bytes(unhashedAddress + addressHash))


    def createFromPrivateKey(self, privateKey):
        if not isinstance(privateKey, SigningKey):
            publicKey = VerifyKey(base58.b58decode(privateKey)[-32:])
            privateKey = base58.b58decode(privateKey)[:-32]
            address = self.createAddress(publicKey)
        else:
            publicKey = privateKey.verify_key
            address = self.createAddress(publicKey)
        return Account(address, publicKey, privateKey)


    def createFromPublicKey(self, publicKey):
        if not isinstance(publicKey, VerifyKey):
            if isinstance(publicKey, bytes):
                decodedPublicKey = VerifyKey(publicKey)
            elif isinstance(publicKey, str):
                decodedPublicKey = VerifyKey(base58.b58decode(publicKey))
            else:
                raise Exception("Unrecognized Public Key format")
            address = self.createAddress(decodedPublicKey)
        else:
            address = self.createAddress(publicKey)
        return Account(address, publicKey)


import os
import struct


from LTO.Account import Account
from LTO import crypto
from LTO.WordList import wordList

import base58

from abc import ABC, abstractmethod

class AccountFactory(ABC):

    def __init__(self, chainId):
        self.chainId = chainId

    @abstractmethod
    def createSignKeys(self, seed):
        pass

    @abstractmethod
    def createAddress(self, publicKey):
        pass

    def create(self):
        return self.createFromSeed(self.generateSeedPhrase())

    def createFromSeed(self, seed, nonce=0):
        privateKey, publicKey = self.createSignKeys(seed, nonce)
        address = self.createAddress(publicKey)
        return Account(address, publicKey, privateKey, seed)

    @abstractmethod
    def createFromPrivateKey(self, privateKey):
        pass

    @abstractmethod
    def createFromPublicKey(self, publicKey):
        pass

    def createWithValues(self, address, publicKey, privateKey, seed=''):
        return Account(address, publicKey, privateKey, seed)

    def assertAccount(self, account, address, publicKey, privateKey, seed):
        if address and account.address != address:
            return False
        if publicKey and account.publicKey != publicKey:
            return False
        if privateKey and account.privateKey != privateKey:
            return False
        return True

    # create the class from the seed
    def generateSeedPhrase(self):
        wordCount = len(wordList)
        words = []
        for i in range(5):
            r = crypto.bytes2str(os.urandom(4))
            x = (ord(r[3])) + (ord(r[2]) << 8) + (ord(r[1]) << 16) + (ord(r[0]) << 24)
            w1 = x % wordCount
            w2 = ((int(x / wordCount) >> 0) + w1) % wordCount
            w3 = ((int((int(x / wordCount) >> 0) / wordCount) >> 0) + w2) % wordCount
            words.append(wordList[w1])
            words.append(wordList[w2])
            words.append(wordList[w3])
        return ' '.join(words)



from nacl.signing import SigningKey, VerifyKey
from LTO import crypto
import base58


class Account(object):

    SODIUM_CRYPTO_SIGN_BYTES = 64
    SODIUM_CRYPTO_SIGN_PUBLICKEYBYTES = 32

    def __init__(self, address, publicKey, privateKey = '', seed='', nonce=0):
        self.address = address
        self.publicKey = publicKey
        self.privateKey = privateKey
        self.seed = seed
        self.nonce = nonce

    def sign(self, message):
        if (self.privateKey == ''):
            raise Exception("Private key not set")
        if isinstance(self.privateKey, SigningKey):
            return base58.b58encode(self.privateKey.sign(message).signature)
        else:
            signature = self.privateKey.sign(message)
            return base58.b58encode(signature)

    def getPublicKey(self):
        if isinstance(self.publicKey, VerifyKey):
            return base58.b58encode(bytes(self.publicKey))
        else:
            return base58.b58encode(self.publicKey.to_string())

    def verifySignature(self, message: str, signature: str, encoding: str = 'base58'):
        if not self.publicKey:
            raise Exception('Unable to verify message; no public sign key')
        rawSignature = crypto.decode(signature, encoding)
        return self.publicKey.verify(message, rawSignature)

    def getNetwork(self):
        return str(base58.b58decode(self.address))[6]



import requests
import json

from LTO.Account import Account

class PublicNode(object):
    def __init__(self, url):
        self.url = url


    def wrapper(self, api, postData='', host='', headers=''):
        if not host:
            host = self.url

        if postData:
            r = requests.post('%s%s' % (host, api), data=postData,
                                headers={'content-type': 'application/json'})
        else:
            r = requests.get('%s%s' % (host, api), headers=headers)


        if r.status_code != 200:
            jsonResp = json.loads(r.text)
            raise Exception('{}'.format(jsonResp['message']))

        r.raise_for_status()

        return r.json()

    def broadcast(self, transaction):
        from LTO import PyCLTO
        data = json.dumps(transaction.toJson())
        response = self.wrapper(api='/transactions/broadcast', postData=data)
        return PyCLTO().fromData(response)

    def getScript(self, scriptSource):
        return self.wrapper('/utils/script/compile', scriptSource)['script'][7:]


    def height(self):
        return self.wrapper('/blocks/height')['height']

    def lastblock(self):
        return self.wrapper('/blocks/last')

    def block(self, n):
        return self.wrapper('/blocks/at/%d' % n)

    def tx(self, id):
        from LTO import PyCLTO
        response = self.wrapper('/transactions/info/%s' % id)
        return PyCLTO().fromData(response)

    def leaseList(self, address):
        return self.wrapper(api='/leasing/active/{}'.format(address))

    def sponsorshipList(self, address):
        return self.wrapper(api='/sponsorship/status/{}'.format(address))

    def balance(self, address):

        if type(address) == Account:
            address = address.address

        try:
            return self.wrapper('/addresses/balance/%s' % address)['balance']
        except:
            return -1

    def transactions(self, limit=100, after='', address = ''):
        return self.wrapper('/transactions/address/%s/limit/%d%s' % (
            address, limit, "" if after == "" else "?after={}".format(after)))




import base58
from LTO import crypto
import struct

from LTO.Transaction import Transaction


class MassTransfer(Transaction):
    DEFAULT_BASE_FEE = 100000000
    TYPE = 11
    DEFAULT_VERSION = 3

    def __init__(self, transfers, attachment=''):
        super().__init__()
        self.transfers = transfers
        self.attachment = attachment
        self.transfersData = ''
        self.baseFee = self.DEFAULT_BASE_FEE
        self.txFee = self.baseFee + int(len(self.transfers) * self.baseFee / 10)
        self.version = self.DEFAULT_VERSION

        if len(self.transfers) > 100:
            raise Exception('Too many recipients')

        self.transfersData = b''
        for i in range(0, len(self.transfers)):
            self.transfersData += base58.b58decode(self.transfers[i]['recipient']) \
                             + struct.pack(">Q", self.transfers[i]['amount'])

    def __toBinaryV1(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\1' +
                base58.b58decode(self.senderPublicKey) +
                struct.pack(">H", len(self.transfers)) +
                self.transfersData +
                struct.pack(">Q", self.timestamp) +
                struct.pack(">Q", self.txFee) +
                struct.pack(">H", len(self.attachment)) +
                crypto.str2bytes(self.attachment))

    def __toBinaryV3(self):
        return (
                self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chainId) +
                struct.pack(">Q", self.timestamp) +
                b'\1' +
                base58.b58decode(self.senderPublicKey) +
                struct.pack(">Q", self.txFee) +
                struct.pack(">H", len(self.transfers)) +

                self.transfersData +

                struct.pack(">H", len(self.attachment)) +
                crypto.str2bytes(self.attachment))

    def toBinary(self):
        if self.version == 1:
            return self.__toBinaryV1()
        elif self.version == 3:
            return self.__toBinaryV3()
        else:
            raise Exception('Incorrect Version')


    def toJson(self):
        return ({
            "type": self.TYPE,
            "version": self.version,
            "sender": self.sender,
            "senderKeyType": "ed25519",
            "senderPublicKey": self.senderPublicKey,
            "fee": self.txFee,
            "timestamp": self.timestamp,
            "proofs": self.proofs,
            "attachment": base58.b58encode(crypto.str2bytes(self.attachment)),
            "transfers": self.transfers
        } | self._sponsorJson())

    @staticmethod
    def fromData(data):
        tx = MassTransfer(transfers='')
        tx.type = data['type']
        tx.version = data['version']
        tx.id = data['id'] if 'id' in data else ''
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.senderKeyType = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.senderPublicKey = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.attachment = data['attachment'] if 'attachment' in data else ''
        tx.transfers = data['transfers']
        tx.height = data['height'] if 'height' in data else ''
        return tx




import base64
import hashlib
import pyblake2
import base58

if bytes == str:  # python2
    str2bytes = lambda s: s
    bytes2str = lambda b: b
    str2list = lambda s: [ord(c) for c in s]
else:  # python3
    str2bytes = lambda s: s.encode('latin-1')
    bytes2str = lambda b: ''.join(map(chr, b))
    str2list = lambda s: [c for c in s]


def sha256(s):
    return hashlib.sha256(str2bytes(s)).digest()


def hashChain(s):
    a = pyblake2.blake2b(s, digest_size=32).digest()
    b = hashlib.sha256(a).digest()
    return ''.join(map(chr, b))

def getNetwork(address):
    # Chain_ID = unpack('Cversion/anetwork', account.address)
    decodedAddress = base58.b58decode(address)
    return str(decodedAddress)[6]

def decode(string, encoding: str):
    if encoding == 'base58':
        return base58.b58decode(string)
    elif encoding == 'base64':
        return base64.b64decode(string)
    elif encoding == 'hex':
        raise Exception('Hexadecimal decoding not yet implemented')
    else:
        raise Exception('Failed to decode')

def encode(string, encoding: str):
    if encoding == 'base58':
        return base58.b58encode(string)
    elif encoding == 'base64':
        return base64.b64encode(string)
    elif encoding == 'hex':
        raise Exception('Hexadecimal encoding not yet implemented')
    else:
        raise Exception('Failed to encode')

def validateAddress(address):
    ADDRESS_VERSION = 1
    ADDRESS_CHECKSUM_LENGTH = 4
    ADDRESS_HASH_LENGTH = 20
    ADDRESS_LENGTH = 1 + 1 + ADDRESS_CHECKSUM_LENGTH + ADDRESS_HASH_LENGTH

    addr = bytes2str(base58.b58decode(address))
    if addr[0] != chr(ADDRESS_VERSION):
        raise Exception('Wrong address version')
    elif len(addr) != ADDRESS_LENGTH:
        raise Exception('Wrong address length')
    elif addr[-ADDRESS_CHECKSUM_LENGTH:] != hashChain(
            str2bytes(addr[:-ADDRESS_CHECKSUM_LENGTH]))[:ADDRESS_CHECKSUM_LENGTH]:
        raise Exception('Wrong address checksum')
    else:
        return True



import base58
from LTO import crypto
from LTO.Transaction import Transaction
import struct

class Anchor(Transaction):
    TYPE = 15
    DEFAULT_ANCHOR_FEE = 35000000
    DEFAULT_VERSION = 3

    def __init__(self, anchor):
        super().__init__()

        self.anchor = anchor
        self.txFee = self.DEFAULT_ANCHOR_FEE
        self.version = self.DEFAULT_VERSION


    def __toBinaryV1(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\1' +
                base58.b58decode(self.senderPublicKey) +
                struct.pack(">H", 1) +
                struct.pack(">H", len(crypto.str2bytes(self.anchor))) +
                crypto.str2bytes(self.anchor) +
                struct.pack(">Q", self.timestamp) +
                struct.pack(">Q", self.txFee))

    def __toBinaryV3(self):
        return (
                self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chainId) +
                struct.pack(">Q", self.timestamp) +
                b'\1' +
                base58.b58decode(self.senderPublicKey) +
                struct.pack(">Q", self.txFee) +
                struct.pack(">H", 1) +
                struct.pack(">H", len(crypto.str2bytes(self.anchor))) +
                crypto.str2bytes(self.anchor)
        )

    def toBinary(self):
        if self.version == 1:
            return self.__toBinaryV1()
        elif self.version == 3:
            return self.__toBinaryV3()
        else:
            raise Exception('Incorrect Version')

    def toJson(self):
        return({
            "type": self.TYPE,
            "version": self.version,
            "sender": self.sender,
            "senderKeyType": "ed25519",
            "senderPublicKey": self.senderPublicKey,
            "fee": self.txFee,
            "timestamp": self.timestamp,
            "anchors": [base58.b58encode(crypto.str2bytes(self.anchor))],
            "proofs": self.proofs
            } | self._sponsorJson())

    @staticmethod
    def fromData(data):
        tx = Anchor(anchor='')
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.senderKeyType = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.senderPublicKey = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.anchors = data['anchors']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''
        return tx



from LTO import crypto
from LTO.Transaction import Transaction
import struct
import base58

class Sponsorship(Transaction):
    TYPE = 18
    DEFAULT_SPONSORSHIP_FEE = 500000000
    DEFAULT_VERSION = 3

    def __init__(self, recipient):
        super().__init__()
        self.recipient = recipient
        crypto.validateAddress(recipient)
        self.txFee = self.DEFAULT_SPONSORSHIP_FEE
        self.version = self.DEFAULT_VERSION

    def __toBinaryV1(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\1' +
                crypto.str2bytes(crypto.getNetwork(self.sender)) +
                base58.b58decode(self.senderPublicKey) +
                base58.b58decode(self.recipient) +
                struct.pack(">Q", self.timestamp) +
                struct.pack(">Q", self.txFee))

    def __toBinaryV3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\1' +
                crypto.str2bytes(self.chainId) +
                struct.pack(">Q", self.timestamp) +
                b'\1' +
                base58.b58decode(self.senderPublicKey) +
                struct.pack(">Q", self.txFee) +
                base58.b58decode(self.recipient)
                )

    def toBinary(self):
        if self.version == 1:
            return self.__toBinaryV1()
        elif self.version == 3:
            return self.__toBinaryV3()
        else:
            raise Exception('Incorrect Version')


    def toJson(self):
        return ({
            "type": self.TYPE,
            "version": self.version,
            "senderKeyType": "ed25519",
            "recipient": self.recipient,
            "sender": self.sender,
            "senderPublicKey": self.senderPublicKey,
            "timestamp": self.timestamp,
            "fee": self.txFee,
            "proofs": self.proofs
        } | self._sponsorJson())

    @staticmethod
    def fromData(data):
        tx = Sponsorship(data['recipient'])
        tx.type = data['type']
        tx.version = data['version']
        tx.id = data['id'] if 'id' in data else ''
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.senderKeyType = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.senderPublicKey = data['senderPublicKey']
        tx.recipient = data['recipient']
        tx.timestamp = data['timestamp']
        tx.fee = data['fee']
        tx.proofs = data['proofs']
        tx.height = data['height'] if 'height' in data else ''
        return tx




import base58
from LTO import crypto
import struct
from LTO.Transaction import Transaction

class Lease(Transaction):
    DEFAULT_LEASE_FEE = 100000000
    TYPE = 8
    DEFAULT_VERSION = 3

    def __init__(self, recipient, amount):
        super().__init__()
        self.amount = amount
        self.recipient = recipient
        # The json response doesn't contain the recipient
        # crypto.validateAddress(recipient)
        self.txFee = self.DEFAULT_LEASE_FEE
        self.version = self.DEFAULT_VERSION
        if self.amount <= 0:
            raise Exception ('Amount must be > 0')


    def __toBinaryV2(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\2' +
                b'\0' +
                base58.b58decode(self.senderPublicKey) +
                base58.b58decode(self.recipient) +
                struct.pack(">Q", self.amount) +
                struct.pack(">Q", self.txFee) +
                struct.pack(">Q", self.timestamp))

    def __toBinaryV3(self):
        return (
                self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chainId) +
                struct.pack(">Q", self.timestamp) +
                b'\1' +
                base58.b58decode(self.senderPublicKey) +
                struct.pack(">Q", self.txFee) +
                base58.b58decode(self.recipient) +
                struct.pack(">Q", self.amount)
        )

    def toBinary(self):
        if self.version == 2:
            return self.__toBinaryV2()
        elif self.version == 3:
            return self.__toBinaryV3()
        else:
            raise Exception('Incorrect Version')

    def toJson(self):
        return ({
            "type": self.TYPE,
            "version": self.version,
            "sender": self.sender,
            "senderKeyType": "ed25519",
            "senderPublicKey": self.senderPublicKey,
            "recipient": self.recipient,
            "amount": self.amount,
            "fee": self.txFee,
            "timestamp": self.timestamp,
            "proofs": self.proofs
        } | self._sponsorJson())

    @staticmethod
    def fromData(data):
        tx = Lease('',1)
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.senderKeyType = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.senderPublicKey = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.recipient = data['recipient']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''
        return tx






from LTO.AccountFactory import AccountFactory
from ecdsa import VerifyingKey, SECP256k1, NIST256p, SigningKey
from ecdsa.util import randrange_from_seed__trytryagain
import base58
from LTO import crypto
from LTO.Account import Account


class AccountECDSA(AccountFactory):

    def __init__(self, chainId, curve='secp256k1'):
        super().__init__(chainId)

        if curve == 'secp256k1':
            self.curve = SECP256k1
        elif curve == 'secp256r1':
            self.curve = NIST256p

    def _MakeKey(self, seed):
        secexp = randrange_from_seed__trytryagain(seed, self.curve.order)
        return SigningKey.from_secret_exponent(secexp, curve=self.curve)

    def createSignKeys(self, seed, nonce=0):
        privateKey = self._MakeKey(seed)
        publicKey = privateKey.verifying_key
        return privateKey, publicKey

    def createAddress(self, publicKey):
        unhashedAddress = chr(1) + str(self.chainId) + crypto.hashChain(publicKey.to_string())[0:20]
        addressHash = crypto.hashChain(crypto.str2bytes(unhashedAddress))[0:4]
        return base58.b58encode(crypto.str2bytes(unhashedAddress + addressHash))

    def createFromPublicKey(self, publicKey):
        if not isinstance(publicKey, VerifyingKey):
            if isinstance(publicKey, bytes):
                publicKey = VerifyingKey.from_string(publicKey, curve=self.curve)
            elif isinstance(publicKey, str):
                publicKey = base58.b58decode(publicKey)
                publicKey = VerifyingKey.from_string(publicKey, curve=self.curve)
            else:
                raise Exception("Unrecognized Public Key format")
        address = self.createAddress(publicKey)
        return Account(address=address, publicKey=publicKey)

    def createFromPrivateKey(self, privateKey):
        if not isinstance(privateKey, SigningKey):
            if isinstance(privateKey, bytes):
                privateKey = SigningKey.from_string(privateKey, curve=self.curve)
            elif isinstance(privateKey, str):
                privateKey = base58.b58decode(privateKey)
                privateKey = SigningKey.from_string(privateKey, curve=self.curve)
            else:
                raise Exception("Unrecognized Private Key format")
        publicKey = privateKey.verifying_key
        address = self.createAddress(publicKey)
        return Account(address=address, publicKey=publicKey, privateKey=privateKey)


from abc import ABC, abstractmethod
from time import time
from LTO.PublicNode import PublicNode
from LTO.Account import Account



class Transaction(ABC):

    def __init__(self):

        self.txFee = 0
        self.timestamp = 0

        self.proofs = []
        self.sender = ''
        self.senderPublicKey = ''
        self.chainId = ''
        self.sponsor = ''
        self.sponsorPublicKey = ''
        self.sponsorKeyType = 'ed25519'


    @abstractmethod
    def toBinary(self):
        pass

    def isSigned(self):
        return len(self.proofs) != 0

    def signWith(self, account: Account):
        if self.timestamp == 0:
            self.timestamp = int(time() * 1000)

        if self.sender == '':
            self.sender = account.address
            self.senderPublicKey = account.getPublicKey()

        self.chainId = account.getNetwork()
        self.proofs.append(account.sign(self.toBinary()))

    def sponsorWith(self, sponsorAccount: Account):
        if not self.isSigned():
            raise Exception('Transaction must be signed first')

        self.sponsor = sponsorAccount.address
        self.sponsorPublicKey = sponsorAccount.publicKey
        self.proofs.append(sponsorAccount.sign(self.toBinary()))

    def broadcastTo(self, node: PublicNode):
        return node.broadcast(self)

    @abstractmethod
    def toJson(self):
        pass

    def _sponsorJson(self):
        if self.sponsor:
            return {"sponsor": self.sponsor,
                    "sponsorPublicKey": self.sponsorPublicKey,
                    "sponsorKeyType": self.sponsorKeyType}
        else:
            return{}

    def __getattr__(self, item):
        return getattr(self, item)



from __future__ import absolute_import, division, print_function, unicode_literals

from LTO.AccountFactory import AccountFactory
from LTO.PublicNode import PublicNode

from LTO.Transactions.Anchor import Anchor
from LTO.Transactions.Lease import Lease
from LTO.Transactions.Association import Association
from LTO.Transactions.CancelLease import CancelLease
from LTO.Transactions.CancelSponsorship import CancelSponsorship
from LTO.Transactions.MassTransfer import MassTransfer
from LTO.Transactions.RevokeAssociation import RevokeAssociation
from LTO.Transactions.SetScript import SetScript
from LTO.Transactions.Sponsorship import Sponsorship
from LTO.Transactions.Transfer import Transfer
from LTO.Accounts.AccountFactoryECDSA import AccountECDSA
from LTO.Accounts.AccountFactoryED25519 import AccountED25519


class PyCLTO:

    def __init__(self, chainId='T'):

        if chainId == 'T':
            self.NODE = PublicNode('https://testnet.lto.network')
        elif chainId == 'L':
            self.NODE = PublicNode('https://nodes.lto.network')
        else:
            self.NODE = ''

        self.chainId = chainId

        self.accountFactories = {
            'ed25519': AccountED25519(chainId),
            'secp256r1': AccountECDSA(chainId, curve='secp256r1'),
            'secp256k1': AccountECDSA(chainId, curve='secp256k1')
        }

    def Account(self, keytype='ed25519', address='', publicKey='', privateKey='', seed='', nonce=0):
        factory = self.accountFactories[keytype]

        if seed:
            account = factory.createFromSeed(seed, nonce)
        elif privateKey:
            account = factory.createFromPrivateKey(privateKey)
        elif publicKey:
            account = factory.createFromPublicKey(publicKey)
        else:
            account = factory.create()

        # We don't have a case for someone who just passes the address
        if not factory.assertAccount(account, address, publicKey, privateKey, seed):
            raise Exception("Accounts info are inconsistent")
        return account

    def getChainId(self):
        return self.chainId

    def fromData(self, data):

        if data['type'] == 4:
            return Transfer(recipient=data['recipient'], amount=data['amount']).fromData(data)
        elif data['type'] == 8:
            return Lease(amount=1, recipient='').fromData(data)
        elif data['type'] == 11:
            return MassTransfer(transfers='').fromData(data)
        elif data['type'] == 15:
            return Anchor(anchor='').fromData(data)
        elif data['type'] == 16:
            return Association(recipient='', associationType='', anchor='').fromData(data)
        elif data['type'] == 17:
            return RevokeAssociation(recipient='', associationType='').fromData(data)
        elif data['type'] == 18:
            return Sponsorship(data['recipient']).fromData(data)
        elif data['type'] == 19:
            return CancelSponsorship(data['recipient']).fromData(data)
        elif data['type'] == 13:
            return SetScript(data['script']).fromData(data)
        elif data['type'] == 9:
            return CancelLease(leaseId='').fromData(data)
        else:
            raise Exception('Incorrect transaction Type')
